/****************************************************
 *													*
 *		Startdatei fr ATMEL-SAMA5D3x-Prozessor		*
 *													*
 *					A. Heinze						*
 *					07.05.13						*
 * zum Teil entnommen aus dem sama5d3x softpack 1.0	*
 *													*
****************************************************/
//------------------------------------------------------------------------------
//	Definitions
//------------------------------------------------------------------------------
					.equ  FREERTOS_COPY_STACK,	1					//wenn das Define im portmacro.h TRUE gesetzt wird, dann hier auf 1 setzen

					.equ AIC,					0xFFFFF000
					.equ AIC_IVR,				0x10
					.equ AIC_FVR,				0x14
					.equ AIC_EOICR,				0x38

					.equ UNDEF_STACK_SIZE,		(16*4)
					.equ ABORT_STACK_SIZE,		(16*4)
					.equ FIQ_STACK_SIZE,		(0)
					.equ IRQ_STACK_SIZE,		((8 * 41 + 15) * 4)	//8 Level  41 Register (lr-irq, spsr, s0-s31, fpscr, fpexc, r0-r3, r12,) und einmalig r0-r14 aus SystemMode -> 1372 Byte

					.equ ARM_MODE_USER,			0x10;
					.equ ARM_MODE_FIQ,			0x11;
					.equ ARM_MODE_IRQ,			0x12;
					.equ ARM_MODE_SVC,			0x13;
					.equ ARM_MODE_MONITOR,		0x16;
					.equ ARM_MODE_ABORT,		0x17;
					.equ ARM_MODE_HYP,			0x1a;
					.equ ARM_MODE_UNDEF,		0x1b;
					.equ ARM_MODE_SYS,			0x1F;

					.equ I_BIT,					0x80;
					.equ F_BIT,					0x40;


					.global CPU_getCPSR
					.global CPU_initVBAR
					.global CPU_prefetchContext
					.global FreeRTOS_restore1stContext
					.global CPU_setupStacks
					.global CPU_spuHandler
					.global CPU_resetVector

					.extern fiqStack
					.extern irqStack
					.extern curNestedIRQ
					.extern CPU_irqVectorStack
					.extern portActAllowTaskSwitch
					.extern portSaveContextPtr
					.extern vPortYieldProcessor

//------------------------------------------------------------------------------
// normale Text Section
//------------------------------------------------------------------------------
					.section .text
					.align 5
					.arm

CPU_getCPSR:		mrs		r0, CPSR
					nop
					nop
					nop
					nop
					bx		lr



CPU_initVBAR:		mcr		p15, 0, r0, c12,  c0, 0		 //WRITE VBAR into Rt
					nop
					nop
					nop
					nop
					bx		lr



/*------------------------------------------------------------------------------
	Stack setzen
/------------------------------------------------------------------------------*/
CPU_setupStacks:	/* FIQ Mode */
					msr		CPSR_c, #ARM_MODE_FIQ | I_BIT | F_BIT
					ldr		sp, =fiqStack
					// IRQ mode
					msr		CPSR_c, #ARM_MODE_IRQ | I_BIT | F_BIT
					ldr		sp, =irqStack

					// Stacksetup - neuer SP in r0
					/* Undefined Instruction Mode */
					msr		CPSR_c, #ARM_MODE_UNDEF | I_BIT | F_BIT
					mov		sp, r0
					sub		r0, sp, #UNDEF_STACK_SIZE
					/* Abort Mode */
					msr		CPSR_c, #ARM_MODE_ABORT | I_BIT | F_BIT
					mov		sp, r0
					sub		r0, sp, #ABORT_STACK_SIZE
					// Supervisor mode
					msr		CPSR_c, #ARM_MODE_SVC | I_BIT | F_BIT
					mov		sp, r0
					//System Mode wird nicht geetzt, das macht das RTOS bei jedem Task
					bx		lr



/*------------------------------------------------------------------------------
	den Context des ersten Tasks holen und den Task ausfhren
/------------------------------------------------------------------------------*/
FreeRTOS_restore1stContext:
					ldr		r0, =portSaveContextPtr;	//den letzten SP holen (beim ersten Restore ist dies der TaskSP)
					ldr		lr,[r0]

					//VPU-Register zurckholen
					ldmia	lr!, {r0}				//vom Stack laden und ins
					vmsr	fpexc, r0				//Floating-Point Exception Register
					ldmia	lr!, {r0}				//vom Stack laden und ins
					vmsr	fpscr, r0				//Floating-Point Status and Control Register
					vldmia	lr!, {s0-s31}			//s0-s31

					ldmia	lr!, {r0}				//SPSR zurckholen
					msr		SPSR_cxsf, r0

					ldmia	lr,{r0-lr}^				//r0-lr des System Modes zurckholen - kein writeBack des SP! ->
					nop
					add		lr, lr, #60				//SP um 15 Register   4Byte hoch

					//setzt PC und schreibt SPSR in CPSR -> damit Modewechsel und IRQ erlaubt (LR des nchsten Modes wurde schon hergestellt)
					ldmia	lr!, {pc}^



//------------------------------------------------------------------------------
// reset
//------------------------------------------------------------------------------
resetHandler:		mov		r0, #0
					MCR		p15, 0, r0, c12,  c0, 0		//VBAR auf 0 setzen
					ldr		pc, [r0, #0]				//und hinspringen







/*------------------------------------------------------------------------------
	Exception vectors
------------------------------------------------------------------------------*/
					.section	.excVectors, "a", %progbits
					.align 5

CPU_resetVector:	ldr		pc, =resetHandler		// Reset */
undefVector:		b		undefVector				// Undefined instruction */
swiVector:			b		swiHandler				// Software interrupt */
prefAbortVector:	b		prefAbortVector			// Prefetch abort */
dataAbortVector:	b		dataAbortVector			// Data abort */
reservedVector:		b		reservedVector			// Reserved for future use */
irqVector:			b		irqHandler				// Interrupt */
fiqVector:			b		fiqHandler				// Fast interrupt */


/*------------------------------------------------------------------------------
	SPU-Handler hier rein
------------------------------------------------------------------------------*/
CPU_spuHandler:		b		CPU_spuHandler


CPU_prefetchContext://der Context ist 216 Byte gross, eine CCL hat 32 Byte
					//also machen wir 7 Prefetches
					pld		[r0, #+0]
					pld		[r0, #+32]
					pld		[r0, #+64]
					pld		[r0, #+96]
					pld		[r0, #+128]
					pld		[r0, #+160]
					pld		[r0, #+192]
					bx		lr


.if FREERTOS_COPY_STACK == 0
/*------------------------------------------------------------------------------
	IRQ-Handler, welcher Nesting erlaubt und den Context-Save/Restore des FreeRTOS untersttzt
	der Context wird direkt auf dem Taskstack abgelegt, um unntiges Kopieren zu vermeiden

	Prinzip:- IRQ-Handler luft zunchst im IRQ-Mode und somit mit eigenem IRQ-Stack und gesperrtem IRQ
			- SP-IRQ sollte mglichst im internen RAM liegen
			- entsprechend der vom FreeRTOS vorgegebenen StackContext-Struktur alle Register sichern, wenn wir aus dem SystemMode kommen (=TaskLevel)
			  LR, R0-R14 (System Mode), SPSR, VPU-Register -> die Adresse nach dem Save wird dem RTOS mitgeteilt
			- zur Nesting-Untersttzung in den Supervisor-Mode wechseln - alle nested IRQs laufen in dem Mode

			- kommen wir nicht aus dem SystemMode, dann werden nur LR und die Scratchregister gesichert
			- zur Nesting-Untersttzung in den Supervisor-Mode wechseln - alle nested IRQs laufen in dem Mode
			- ab jetz knnte ein hher priorisierter IRQ reinhaken
			- ISR ausfhren
			- wieder in den IRQ-Mode wechseln (mit gesperrtem IRQ)
			- WICHTIG!: - checken, ob wir kurz vor der Rckkehr zum SystemMode stehen (bzw. zum niedrigsten IRQ-Level)
						- dies geschieht durch Prfen des SPSR
						- wenn ja, dann dem RTOS Bescheid geben, dass der derzeitig eingestellte Task drankommt
						  (ausser in der Zeit, des zwischen soeben erfolgten Wechsels in IRQ Mode und Verlassen kommt ein IRQ)
						- dazu incrementieren wir einfach eine Variable
						- das RTOS prft dann beim nchsten SAVE-CONTEXT, ob das Increment erfolgte, wenn ja, dann wird geswitched
						- wenn nein, dann kein Switch, da zum Task noch nicht zurckgekehrt wurde und somit die Gefahr besteht,
						  dass IRQs die Task nicht nur blockieren, sondern sie nie drankommen, da sie quasi "berswitched" werden
			- IRQ-Ack durchfhren (EOICR schreiben)
			- SPSR, R0-R14 (System) und LR wiederherstellen
			- und raus dem IRQ-Handler

			- Ergnzung: - zur Untersttzung der Abfrage des aktuellen IRQ-Levels wird beim Eintritt in die ISR eine Variable incrementiert
						 - und in ein Array mit dem Index dieser Variable der aktuelle ISR-Vector abgelegt, dass man spter an Hand des Vectors die Prio ermitteln kann
						 - am Ende der ISR wird wieder decrementiert
/------------------------------------------------------------------------------*/
irqHandler:			//wenn Scheduler des RTOS im FIQ luft, dann auch den sperren
					cpsid	f
					// entsprechend der vom FreeRTOS vorgegebenen StackContext-Struktur alle Register sichern
					// LR, R0-R14 (System Mode), SPSR, CritNesting-Variable
					// dazu vorher
					sub		lr, lr, #4
					stmdb	sp!, {lr}				//LR_irq sichern
					mrs		lr, SPSR				//SPSR holen
					and		lr, lr, #ARM_MODE_SYS	//Mode maskieren (=ARM_MODE_SYS) und in R1 ablegen
					cmp		lr, #ARM_MODE_SYS		//war letzter Mode SystemMode?
					bne		_notSysModeIRQ1			//wenn nich, dann weiter

					//sonst: letzter Mode war das TaskLevel -> alle weiteren Register AUF DEM TASKSTACK sichern
					//dazu brauchen wir erst einmal den SP des Tasks
					ldmia	sp!, {lr}				//LR_irq wiederherstellen (dieser muss als erstes auf den TaskStack)
					stmdb	sp!, {r0}				//r0 freimachen
					stmdb	sp,{sp}^				//sp_usr auf den Stack - kein writeBack des SP! ->
					nop
					sub		sp, sp, #4				//SP um 4Byte runter
					ldmia	sp!, {r0}				//nun liegt in R0 der Stackpointer des Tasks
					stmdb	r0!, {lr}				//zuerst lr_irq sichern
					mov		lr, r0					//alles weitere machen wir mit dem LR
					ldmia	sp!, {r0}				//r0 wiederherstellen
					stmdb	lr,{r0-lr}^				//r0-lr des System Modes sichern - kein writeBack des LR! ->
					nop
					sub		lr, lr, #60				//SP um 15 Register   4Byte runter
					mrs		r0, SPSR				//SPSR holen
					stmdb	lr!, {r0}				//SPSR sichern
					//und noch alles von der VFU sichern
					vstmdb	lr!, {s0-s31}			//s0-s31
					vmrs	r0, fpscr				//Floating-Point Status and Control Register
					stmdb	lr!, {r0}				//sichern
					vmrs	r0, fpexc				//Floating-Point Exception Register
					stmdb	lr!, {r0}				//sichern

					//dem RTOS Bescheid geben, wo wir den ganzen Kram hingeschrieben haben
					ldr		r0, =portSaveContextPtr;
					str		lr, [r0]

					// IVR in r0 lesen + schreiben, dass der AIC den Protect Mode fr Debugzwecke untersttzt
					ldr		lr, =AIC
					ldr		r0, [r14, #AIC_IVR]		//IRQ-Vector in r0 lesen (lr_irq geht nicht, da spter beim Wechsel in SV-Mode lr von lr_irq zu lr_svc gewechselt wird)
					str		r0, [r14, #AIC_IVR]		//Protect-Mode ein

					//Zhler fr das aktuellen IRQ-Nest hochzhlen
					ldr		r1, =curNestedIRQ;
					ldr		r2, [r1]
					add		r2, r2, #1
					str		r2, [r1]

					//IRQ-Vector im IRQNestArray sichern (vlt. besser die Prio sichern???)
					ldr		r1, =CPU_irqVectorStack;
					str		r0, [r1, r2, LSL #2]

					// und auf gehts zur ISR... IRQ-MOde -> SVC Mode u. ISR ausfhren -> zurck in IRQ-Mode
					// Wechsel in den Supervisor-Mode (ab hier IRQ/FIQ erlaubt)
					msr		CPSR_c, #ARM_MODE_SVC
					//den Stack auf 8 Byte alignen und noch das LR_sv des SV-Modes sichern
					and		r1, sp, #4
					sub		sp, sp, r1
					stmdb	sp!, {r1, lr}
					//ISR ausfhren
					blx		r0
					//LR_sv des SV-Modes zurckholen und den Stack zurckbiegen
					ldmia	sp!, {r1, lr}
					add		sp, sp, r1
					//LR_sv  zurckholen und Wechsel in den IRQ-Mode (mit gesperrtem IRQ/FIQ)
					msr		CPSR_c, #ARM_MODE_IRQ | I_BIT| F_BIT

//					mov		r0, #0x400				// Debug Pin0 High ODER
//					str		r0, [r0, #-(0x7d0 + 0x400)]
//					mov		r0, #0x800				// Debug Pin1 High
//					str		r0, [r0, #-(0x7d0 + 0x800)]

					//Zhler fr das aktuellen IRQ-Nest runterzhlen
					ldr		r0, =curNestedIRQ;
					ldr		r1, [r0]
					sub		r1, r1, #1
					str		r1, [r0]

					//den Zeiger auf den aktuellen Context in LR holen
					ldr		r0, =portSaveContextPtr;
					ldr		lr, [r0]

					//und nun den Context ab dort holen
					//VPU-Register zurckholen
					ldmia	lr!, {r0}				//vom Taskstack laden und ins
					vmsr	fpexc, r0				//Floating-Point Exception Register
					ldmia	lr!, {r0}				//vom Taskstack laden und ins
					vmsr	fpscr, r0				//Floating-Point Status and Control Register
					vldmia	lr!, {s0-s31}			//s0-s31
					ldmia	lr!, {r0}				//SPSR zurckholen
					msr		SPSR_cxsf, r0

					//dem RTOS Bescheid geben, dass wieder ein ContextSwitch erfolgen darf
					ldr		r0, =portActAllowTaskSwitch
					ldr		r1, [r0]
					add		r1, r1, #1
					str		r1, [r0]
					ldmia	lr,{r0-lr}^				//r0-lr des System Modes zurckholen - kein writeBack des SP! ->
					nop
					ldr		lr, [lr, #+60]			//lr laden (15 Register   4Byte hher)
					stmdb	sp!, {lr}				//und auf dem Stack sichern

//					mov		lr, #0x400				// Debug Pin0 Low ODER
//					str		lr, [lr, #-(0x7cc + 0x400)]
//					mov		lr, #0x800				// Debug Pin1 Low
//					str		lr, [lr, #-(0x7cc + 0x800)]

					// Ack IRQ
					ldr		lr, =AIC
					str		lr, [lr, #AIC_EOICR]

					//setzt PC und schreibt SPSR in CPSR -> damit Modewechsel und IRQ/FIQ erlaubt (LR des nchsten Modes wurde schon hergestellt)
					ldmia	sp!, {pc}^
					//Ende der ISR fr den Systemmode


					//ISR, wenn wir schon beim Nesten sind
					//LR ist schon auf dem Stack, wir knnen also gleich damit arbeiten
_notSysModeIRQ1:	mrs		lr, SPSR				//SPSR holen
					stmdb	sp!, {lr}				//SPSR sichern
					//alles von der VFU sichern
					vstmdb	sp!, {s0-s31}			//s0-s31
					vmrs	lr, fpscr				//Floating-Point Status and Control Register
					stmdb	sp!, {lr}				//sichern
					vmrs	lr, fpexc				//Floating-Point Exception Register
					stmdb	sp!, {lr}				//sichern

					// Scratchregister sichern (r0-r3, r12) und
					// IVR in r0 lesen + schreiben, dass der AIC den Protect Mode fr Debugzwecke untersttzt
					stmdb	sp!, {r0-r3, r12}
					ldr		lr, =AIC
					ldr		r0, [r14, #AIC_IVR]		//IRQ-Vector in r0 lesen (lr_irq geht nicht, da spter beim Wechsel in SV-Mode lr von lr_irq zu lr_svc gewechselt wird)
					str		r0, [r14, #AIC_IVR]		//Protect-Mode ein

					//Zhler fr das aktuellen IRQ-Nest hochzhlen
					ldr		r1, =curNestedIRQ;
					ldr		r2, [r1]
					add		r2, r2, #1
					str		r2, [r1]

					//IRQ-Vector im IRQNestArray sichern (vlt. besser die Prio sichern???)
					ldr		r1, =CPU_irqVectorStack;
					str		r0, [r1, r2, LSL #2]

					// und auf gehts zur ISR... IRQ-MOde -> SVC Mode u. ISR ausfhren -> zurck in IRQ-Mode
					// Wechsel in den Supervisor-Mode (ab hier IRQ/FIQ erlaubt)
					msr		CPSR_c, #ARM_MODE_SVC
					//den Stack auf 8 Byte alignen und noch das LR_sv des SV-Modes sichern
					and		r1, sp, #4
					sub		sp, sp, r1
					stmdb	sp!, {r1, lr}
					//ISR ausfhren
					blx		r0
					//LR_sv des SV-Modes zurckholen und den Stack zurckbiegen
					ldmia	sp!, {r1, lr}
					add		sp, sp, r1
					//LR_sv  zurckholen und Wechsel in den IRQ-Mode (mit gesperrtem IRQ/FIQ)
					msr		CPSR_c, #ARM_MODE_IRQ | I_BIT| F_BIT

					//Zhler fr das aktuellen IRQ-Nest runterzhlen
					ldr		r0, =curNestedIRQ;
					ldr		r1, [r0]
					sub		r1, r1, #1
					str		r1, [r0]
					// Scratchregister zurckholen (r0-r3, r12)
					ldmia	sp!, {r0-r3, r12}

					//VPU-Register zurckholen
					ldmia	sp!, {lr}				//vom Stack laden und ins
					vmsr	fpexc, lr				//Floating-Point Exception Register
					ldmia	sp!, {lr}				//vom Stack laden und ins
					vmsr	fpscr, lr				//Floating-Point Status and Control Register
					vldmia	sp!, {s0-s31}			//s0-s31

					//SPSR zurckholen
					ldmia	sp!, {lr}
					msr		SPSR_cxsf, lr

					// Ack IRQ
					ldr		lr, =AIC
					str		lr, [lr, #AIC_EOICR]

					//setzt PC und schreibt SPSR in CPSR -> damit Modewechsel und IRQ/FIQ erlaubt (LR des nchsten Modes wurde schon hergestellt)
					ldmia	sp!, {pc}^



.else
/*------------------------------------------------------------------------------
	IRQ-Handler, welcher Nesting erlaubt und den Context-Save/Restore des FreeRTOS untersttzt

	Prinzip:- IRQ-Handler luft zunchst im IRQ-Mode und somit mit eigenem IRQ-Stack und gesperrtem IRQ
			- SP-IRQ sollte mglichst im internen RAM liegen
			- entsprechend der vom FreeRTOS vorgegebenen StackContext-Struktur alle Register sichern, wenn wir aus dem SystemMode kommen (=TaskLevel)
			  LR, R0-R14 (System Mode), SPSR -> die Adresse des SPs nach dem Save wird dem RTOS mitgeteilt
			- kommen wir nicht aus dem SystemMode, dann werden nur LR,die Scratch- und VPU-Register gesichert
			- zur Nesting-Untersttzung in den Supervisor-Mode wechseln - alle nested IRQs laufen in dem Mode
			- ab jetz knnte ein hher priorisierter IRQ reinhaken
			- ISR ausfhren
			- wieder in den IRQ-Mode wechseln (mit gesperrtem IRQ)
			- WICHTIG!: - checken, ob wir kurz vor der Rckkehr zum SystemMode stehen (bzw. zum niedrigsten IRQ-Level)
						- dies geschieht durch Prfen des SPSR
						- wenn ja, dann dem RTOS Bescheid geben, dass der derzeitig eingestellte Task drankommt
						  (ausser in der Zeit, des zwischen soeben erfolgten Wechsels in IRQ Mode und Verlassen kommt ein IRQ)
						- dazu incrementieren wir einfach eine Variable
						- das RTOS prft dann beim nchsten SAVE-CONTEXT, ob das Increment erfolgte, wenn ja, dann wird geswitched
						- wenn nein, dann kein Switch, da zum Task noch nicht zurckgekehrt wurde und somit die Gefahr besteht,
						  dass IRQs die Task nicht nur blockieren, sondern sie nie drankommen, da sie quasi "berswitched" werden
			- IRQ-Ack durchfhren (EOICR schreiben)
			- VPU-Register, SPSR, R0-R14 (System) und LR wiederherstellen
			- und raus dem IRQ-Handler

			- Ergnzung: - zur Untersttzung der Abfrage des aktuellen IRQ-Levels wird beim Eintritt in die ISR eine Variable incrementiert
						 - und in ein Array mit dem Index dieser Variable der aktuelle ISR-Vector abgelegt, dass man spter an Hand des Vectors die Prio ermitteln kann
						 - am Ende der ISR wird wieder decrementiert
/------------------------------------------------------------------------------*/
irqHandler:			//wenn Scheduler des RTOS im FIQ luft, dann auch den sperren
					cpsid	f
					// entsprechend der vom FreeRTOS vorgegebenen StackContext-Struktur alle Register sichern
					// LR, R0-R14 (System Mode), SPSR, CritNesting-Variable
					sub		lr, lr, #4
					stmdb	sp!, {lr}				//LR_irq sichern
					mrs		lr, SPSR				//SPSR holen
					and		lr, lr, #ARM_MODE_SYS	//Mode maskieren (=ARM_MODE_SYS) und in R1 ablegen
					cmp		lr, #ARM_MODE_SYS		//war letzter Mode SystemMode?
					bne		_notSysModeIRQ1			//wenn nich, dann weiter

					//sonst: letzter Mode war das TaskLevel -> alle weiteren Register sichern
					stmdb	sp,{r0-lr}^				//r0-lr des System Modes sichern - kein writeBack des SP! ->
					nop
					sub		sp, sp, #60				//SP um 15 Register   4Byte runter
					ldr		r0, =portSaveContextPtr;	//dem RTOS Bescheid geben, wo wir sind (eigentlich ist dies immer dieselbe Adresse, ausser fr SWI)
					str		sp, [r0]

				//	mov		r0, #0x400				// Debug Pin0 High ODER
				//	mov		r0, #0x1000				// Debug Pin3 High
				//	mvn		r1, #0x0
				//	str		r0, [r1, #-0x7CF]

//					ldr		r0, =dbg_irqStacks;
//					blx		r0

_notSysModeIRQ1:	mrs		lr, SPSR				//SPSR holen
					stmdb	sp!, {lr}				//SPSR sichern
					//und noch alles von der VFU sichern
					vstmdb	sp!, {s0-s31}			//s0-s31
					vmrs	lr, fpscr				//Floating-Point Status and Control Register
					stmdb	sp!, {lr}				//sichern
					vmrs	lr, fpexc				//Floating-Point Exception Register
					stmdb	sp!, {lr}				//sichern

					// Scratchregister sichern (r0-r3, r12) und
					// IVR in r0 lesen + schreiben, dass der AIC den Protect Mode fr Debugzwecke untersttzt
					stmdb	sp!, {r0-r3, r12}
					ldr		lr, =AIC
					ldr		r0, [r14, #AIC_IVR]		//IRQ-Vector in r0 lesen (lr_irq geht nicht, da spter beim Wechsel in SV-Mode lr von lr_irq zu lr_svc gewechselt wird)
					str		r0, [r14, #AIC_IVR]		//Protect-Mode ein

					//Zhler fr das aktuellen IRQ-Nest hochzhlen
					ldr		r1, =curNestedIRQ;
					ldr		r2, [r1]
					add		r2, r2, #1
					str		r2, [r1]

					//IRQ-Vector im IRQNestArray sichern (vlt. besser die Prio sichern???)
					ldr		r1, =CPU_irqVectorStack;
					str		r0, [r1, r2, LSL #2]

					// und auf gehts zur ISR... IRQ-MOde -> SVC Mode u. ISR ausfhren -> zurck in IRQ-Mode
					// Wechsel in den Supervisor-Mode (ab hier IRQ/FIQ erlaubt)
					msr		CPSR_c, #ARM_MODE_SVC
					//den Stack auf 8 Byte alignen und noch das LR_sv des SV-Modes sichern
					and		r1, sp, #4
					sub		sp, sp, r1
					stmdb	sp!, {r1, lr}
					//ISR ausfhren
					blx		r0
					//LR_sv des SV-Modes zurckholen und den Stack zurckbiegen
					ldmia	sp!, {r1, lr}
					add		sp, sp, r1
					//LR_sv  zurckholen und Wechsel in den IRQ-Mode (mit gesperrtem IRQ/FIQ)
					msr		CPSR_c, #ARM_MODE_IRQ | I_BIT| F_BIT

					//Zhler fr das aktuellen IRQ-Nest runterzhlen
					ldr		r0, =curNestedIRQ;
					ldr		r1, [r0]
					sub		r1, r1, #1
					str		r1, [r0]
					// Scratchregister zurckholen (r0-r3, r12)
					ldmia	sp!, {r0-r3, r12}

					//VPU-Register zurckholen
					ldmia	sp!, {lr}				//vom Stack laden und ins
					vmsr	fpexc, lr				//Floating-Point Exception Register
					ldmia	sp!, {lr}				//vom Stack laden und ins
					vmsr	fpscr, lr				//Floating-Point Status and Control Register
					vldmia	sp!, {s0-s31}			//s0-s31

					// Rcksprunglevel = Tasklevel?
					ldmia	sp!, {lr}				//SPSR zurckholen
					msr		SPSR_cxsf, lr
					and		lr, lr, #ARM_MODE_SYS	//Mode maskieren (=ARM_MODE_SYS) und in lr ablegen
					cmp		lr, #ARM_MODE_SYS		//ist nchster Mode SystemMode?
					bne		_notSysModeIRQ2			//wenn nich, dann weiter

//					ldr		r0, =checkStack;
//					blx		r0

				//	mov		r0, #0x400				// Debug Pin0 Low ODER
				//	mov		r0, #0x1000				// Debug Pin3 Low
				//	mvn		r1, #0x0
				//	str		r0, [r1, #-0x7CB]

					//wenn ja, dann RTOS Bescheid geben, dass wieder ein ContextSwitch erfolgen darf
					ldr		r0, =portActAllowTaskSwitch
					ldr		r1, [r0]
					add		r1, r1, #1
					str		r1, [r0]
					ldmia	sp,{r0-lr}^				//r0-lr des System Modes zurckholen - kein writeBack des SP! ->
					nop
					add		sp, sp, #60				//SP um 15 Register   4Byte hoch

_notSysModeIRQ2:	// Ack IRQ
					ldr		lr, =AIC
					str		lr, [lr, #AIC_EOICR]

					//setzt PC und schreibt SPSR in CPSR -> damit Modewechsel und IRQ/FIQ erlaubt (LR des nchsten Modes wurde schon hergestellt)
					ldmia	sp!, {pc}^
.endif



/*------------------------------------------------------------------------------
	FIQ-Handler, welcher Nesting erlaubt und den Context-Save/Restore des FreeRTOS untersttzt (Scheduler luft als einzige ISR im FIQ)

	auf Grund eines noch unbekannten Problems, welches den Stackt zerrdelt, luft der Scheduler jetz im IRQ-Level 7 und fiq ist TOT
/------------------------------------------------------------------------------*/
fiqHandler:			b		fiqHandler




/*------------------------------------------------------------------------------
	SWI-Handler hier rein

	der swi Aufruf hat mit dem RTOS nicht wie gewnscht funktioniert, ein portYield wird nun ber einen generierten IRQ ausgelst

------------------------------------------------------------------------------*/
swiHandler:			b		swiHandler

