/*
    FreeRTOS V7.4.2 - Copyright (C) 2013 Real Time Engineers Ltd.

    FEATURES AND PORTS ARE ADDED TO FREERTOS ALL THE TIME.  PLEASE VISIT
    http://www.FreeRTOS.org TO ENSURE YOU ARE USING THE LATEST VERSION.

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.

    >>>>>>NOTE<<<<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.

    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details. You should have received a copy of the GNU General Public License
    and the FreeRTOS license exception along with FreeRTOS; if not it can be
    viewed here: http://www.freertos.org/a00114.html and also obtained by
    writing to Real Time Engineers Ltd., contact details for whom are available
    on the FreeRTOS WEB site.

    1 tab == 4 spaces!

    ***************************************************************************
     *                                                                       *
     *    Having a problem?  Start by reading the FAQ "My application does   *
     *    not run, what could be wrong?"                                     *
     *                                                                       *
     *    http://www.FreeRTOS.org/FAQHelp.html                               *
     *                                                                       *
    ***************************************************************************


    http://www.FreeRTOS.org - Documentation, books, training, latest versions,
    license and Real Time Engineers Ltd. contact details.

    http://www.FreeRTOS.org/plus - A selection of FreeRTOS ecosystem products,
    including FreeRTOS+Trace - an indispensable productivity tool, and our new
    fully thread aware and reentrant UDP/IP stack.

    http://www.OpenRTOS.com - Real Time Engineers ltd license FreeRTOS to High
    Integrity Systems, who sell the code with commercial support,
    indemnification and middleware, under the OpenRTOS brand.

    http://www.SafeRTOS.com - High Integrity Systems also provide a safety
    engineered and independently SIL3 certified version for use in safety and
    mission critical applications that require provable dependability.
*/


/*-----------------------------------------------------------
 * Components that can be compiled to either ARM or THUMB mode are
 * contained in port.c  The ISR routines, which can only be compiled
 * to ARM mode, are contained in this file.
 *----------------------------------------------------------*/

/*
	Changes from V3.2.4

	+ The assembler statements are now included in a single asm block rather
	  than each line having its own asm block.
*/

/*
 * Modified by A.Heinze and M.Bhme/HAMEG Instruments GmbH
 * - Atmel SAMA5D31 registers
 * - per task reent structs
 * - stack checksum
 * - logging of task switches
 * - user tick func
 */


/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"

#include <stddef.h>
#include <reent.h>
#include <string.h>

#include <debug.h>
#include <IO_SAMA5D3x.h>
#include <CPU_SAMA5D3_1.h>


/* Constants required to handle critical sections. */
#define portNO_CRITICAL_NESTING		( ( unsigned long ) 0 )
static unsigned long ulCriticalNesting = 0;


#ifndef FREERTOS_DEBUG_PIN_SYSTICK
#define FREERTOS_DEBUG_PIN_SYSTICK			255
#endif

#ifndef FREERTOS_DEBUG_PIN_SAVE_CONTEXT
#define FREERTOS_DEBUG_PIN_SAVE_CONTEXT		255
#endif

#ifndef FREERTOS_DEBUG_PIN_RESTORE_CONTEXT
#define FREERTOS_DEBUG_PIN_RESTORE_CONTEXT	255
#endif

#ifndef FREERTOS_DEBUG_PIN_SAVE_CONTEXT
#define FREERTOS_DEBUG_PIN_SAVE_CONTEXT		255
#endif

#ifndef FREERTOS_DEBUG_PIN_RESTORE_CONTEXT
#define FREERTOS_DEBUG_PIN_RESTORE_CONTEXT	255
#endif

#ifndef FREERTOS_DEBUG_PIN_PORTYIELD
#define FREERTOS_DEBUG_PIN_PORTYIELD		255
#endif

#ifndef FREERTOS_DEBUG_PIN_PORTYIELD_ISR
#define FREERTOS_DEBUG_PIN_PORTYIELD_ISR	255
#endif

#ifndef FREERTOS_DEBUG_PIN_PORTYIELD_PROC
#define FREERTOS_DEBUG_PIN_PORTYIELD_PROC	255
#endif

#ifndef FREERTOS_DEBUG_PIN_ENTERCRITICAL
#define FREERTOS_DEBUG_PIN_ENTERCRITICAL	255
#endif

#ifndef FREERTOS_DEBUG_PIN_EXITCRITICAL
#define FREERTOS_DEBUG_PIN_EXITCRITICAL		255
#endif


/*-----------------------------------------------------------*/

// siehe FreeRTOS tasks.c
extern portSTACK_TYPE *pxCurrentTCB;

portSTACK_TYPE *restoredTCB;


portSTACK_TYPE *portSaveContextPtr = NULL;
unsigned int	portActAllowTaskSwitch, portSavedAllowTaskSwitch, portSchedulerSuspended;

unsigned int	CPU_irqVectorStack[CPU_IRQ_PRIOR_NO];
unsigned int	curNestedIRQ;

#if FREERTOS_REENT_CHANGE_POINTER
// Zeiger auf das reent-struct im Stack des aktuellen Task
struct _reent	*portCurrentTaskReentPtr = NULL;
#endif

/*-----------------------------------------------------------*/
// liegen in der _isrPort.s
extern void FreeRTOS_restore1stContext( void );
extern void CPU_prefetchContext(portSTACK_TYPE *context);


/*
 * The scheduler can only be started from ARM mode, hence the inclusion of this
 * function here.
 */
void vPortISRStartFirstTask( void );


#if defined DEBUG_RTOS_STACK_CHECK_CRC
// Prfsumme ber Task Stack
extern portStack_TYPE xPortCalcStackCrc (const portStack_TYPE *pxTopOfStack)
#endif

/*-----------------------------------------------------------*/

#if defined DEBUG_RTOS_CONTEXT_LOG_LENGTH && DEBUG_RTOS_CONTEXT_LOG_LENGTH > 0
//#define DBG_IRQSTACKS_NO	DEBUG_RTOS_CONTEXT_LOG_LENGTH
#define DBG_RTOSSTACKS_NO	DEBUG_RTOS_CONTEXT_LOG_LENGTH

#elif (DEBUG_EDITOR == DEBUG_EDITOR_AHC) || (DEBUG_EDITOR == DEBUG_EDITOR_MBC)
#define _DBG_IRQSTACKS_NO	1024
#define _DBG_RTOSSTACKS_NO	1024

#endif

#ifdef DBG_IRQSTACKS_NO
#warning "DBG_IRQSTACKS_NO"

unsigned int dbgIrqStackCnt;

typedef struct irqStackDump_struct
{
	portSTACK_TYPE	*sp_sys;
	void			*pc_sys;
	void			*lr_sys;
	//Die Register lassen wir erstma wech
}irqStackDump_type;

irqStackDump_type dbgIrqStacks[DBG_IRQSTACKS_NO];

void dbg_irqStacks(void)
{
	if (dbgIrqStackCnt >= DBG_IRQSTACKS_NO)
		dbgIrqStackCnt = 0;

	dbgIrqStacks[dbgIrqStackCnt].sp_sys = (portSTACK_TYPE*)portSaveContextPtr[REG_SYS_SP - REG_SAVECONTEXT_PTR];
	dbgIrqStacks[dbgIrqStackCnt].pc_sys = (void*)portSaveContextPtr[REG_IRQ_LR - REG_SAVECONTEXT_PTR];
	dbgIrqStacks[dbgIrqStackCnt].lr_sys = (void*)portSaveContextPtr[REG_SYS_LR - REG_SAVECONTEXT_PTR];
	dbgIrqStackCnt++;
	dbgIrqStacks[dbgIrqStackCnt].sp_sys = (portSTACK_TYPE*)(0);
	dbgIrqStacks[dbgIrqStackCnt].pc_sys = (void*)(0);
	dbgIrqStacks[dbgIrqStackCnt].lr_sys = (void*)(0);
}
#endif

#ifdef DBG_RTOSSTACKS_NO
#warning "DBG_RTOSSTACKS_NO"

#if ( configGENERATE_RUN_TIME_STATS == 1 )
#warning "configGENERATE_RUN_TIME_STATS"
#endif

unsigned int dbgRtosStackCnt;

/*
 * Task control block.  A task control block (TCB) is allocated for each task,
 * and stores task state information, including a pointer to the task's context
 * (the task's run time environment, including register values)
 */
typedef struct dbgTskTaskControlBlock
{
	volatile portSTACK_TYPE	*pxTopOfStack;		/*< Points to the location of the last item placed on the tasks stack.  THIS MUST BE THE FIRST MEMBER OF THE TCB STRUCT. */

	#if ( portUSING_MPU_WRAPPERS == 1 )
		xMPU_SETTINGS xMPUSettings;				/*< The MPU settings are defined as part of the port layer.  THIS MUST BE THE SECOND MEMBER OF THE TCB STRUCT. */
	#endif

	xListItem				xGenericListItem;		/*< The list that the state list item of a task is reference from denotes the state of that task (Ready, Blocked, Suspended ). */
	xListItem				xEventListItem;		/*< Used to reference a task from an event list. */
	unsigned portBASE_TYPE	uxPriority;			/*< The priority of the task.  0 is the lowest priority. */
	portSTACK_TYPE			*pxStack;			/*< Points to the start of the stack. */
	signed char				pcTaskName[ configMAX_TASK_NAME_LEN ];/*< Descriptive name given to the task when created.  Facilitates debugging only. */

	#if ( portSTACK_GROWTH > 0 )
		portSTACK_TYPE *pxEndOfStack;			/*< Points to the end of the stack on architectures where the stack grows up from low memory. */
	#endif

	#if ( portCRITICAL_NESTING_IN_TCB == 1 )
		unsigned portBASE_TYPE uxCriticalNesting; /*< Holds the critical section nesting depth for ports that do not maintain their own count in the port layer. */
	#endif

	#if ( configUSE_TRACE_FACILITY == 1 )
		unsigned portBASE_TYPE	uxTCBNumber;	/*< Stores a number that increments each time a TCB is created.  It allows debuggers to determine when a task has been deleted and then recreated. */
		unsigned portBASE_TYPE  uxTaskNumber;	/*< Stores a number specifically for use by third party trace code. */
	#endif

	#if ( configUSE_MUTEXES == 1 )
		unsigned portBASE_TYPE uxBasePriority;	/*< The priority last assigned to the task - used by the priority inheritance mechanism. */
	#endif

	#if ( configUSE_APPLICATION_TASK_TAG == 1 )
		pdTASK_HOOK_CODE pxTaskTag;
	#endif

	#if ( configGENERATE_RUN_TIME_STATS == 1 )
		unsigned long ulRunTimeCounter;			/*< Stores the amount of time the task has spent in the Running state. */
	#endif

} dbgTskTCB;


typedef struct rtosStackDump_struct
{
	portSTACK_TYPE	*sp_sys;		/*!< Stack Pointer */
	portSTACK_TYPE	*context;
	portSTACK_TYPE	*TopOfStack;
//	portSTACK_TYPE	*pxCurrentTCB;
	dbgTskTCB		*pxCurrentTCB;
	unsigned int	critNest;
	unsigned int	ticks;
#ifdef DBG_IRQSTACKS_NO
	unsigned int	curDbgIrqStackCnt;
#endif
	unsigned int	uxPriority;		/*< The priority of the task.  0 is the lowest priority. */
#if ( configGENERATE_RUN_TIME_STATS == 1 )
	unsigned long	ulRunTimeCounter;			/*< Stores the amount of time the task has spent in the Running state. */
#endif

#if 0	// die Anzeige der TaskName dauert ewig im Trace32
	unsigned int		lr_sys;			/*!< Link Register als Zahl, damit die FTAB im Trace32 schner aussieht */
	unsigned int		pc_sys;			/*!< Program Counter als Zahl, damit die FTAB im Trace32 schner aussieht*/
	const signed char	*pcTaskName;	/*!< Descriptive name given to the task when created.  Facilitates debugging only. */
#elif 1
	unsigned int		pc_sys;			/*!< Program Counter als Zahl, damit die FTAB im Trace32 schner aussieht*/
	void				*lr_sys;		/*!< Link Register als letztes fr FTAB */
#else
	unsigned int		lr_sys;			/*!< Link Register als Zahl, damit die FTAB im Trace32 schner aussieht */
	void				*pc_sys;		/*!< Program Counter als letztes fr FTAB */
#endif
	//Die Register lassen wir erstma wech
} rtosStackDump_type;

rtosStackDump_type dbgRtosStacks[DBG_RTOSSTACKS_NO];

#endif	// DBG_RTOSSTACKS_NO

/*-----------------------------------------------------------*/


//beide CONTEXT-Funcs mssen kein asm mehr sein, da sie vom IRQ-Handler gerufen werden
//die yieldFunc, welche mit SVC aus SystemMode aufgerufen wird, nimmt weiterhin die alten originalen Macros

void portRESTORE_CONTEXT( void )
{
	//wird nur aufgerufen, wenn Contextswitch efolgte
	portSTACK_TYPE *taskSp;

	//nur restoren, wenn du darfst
	if ((portSchedulerSuspended == FALSE) && (portSaveContextPtr != NULL))
	{
		DEBUG_PIN_HIGH_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_RESTORE_CONTEXT);
		NDEBUG_PIN_HIGH_ (DEBUG_EDITOR_MBC, FREERTOS_DEBUG_PIN_RESTORE_CONTEXT);

		taskSp = (portSTACK_TYPE*)(*pxCurrentTCB);

#if defined DEBUG_RTOS_STACK_CHECK_CRC
		{
			portSTACK_TYPE crc;

			// Prfsumme ber den aktuellen Stackinhalt, muss mit dem gespeicherten Wert bereinstimmen
			crc = xPortCalcStackCrc (&taskSp[REG_STACK_CRC + 1]);
			assert (crc == taskSp[REG_STACK_CRC]);
		}
#endif

		if (ulCriticalNesting > 0)
		{
			// zum Brechen
			asm volatile ("nop");
		}
		ulCriticalNesting = taskSp[REG_CRITNEST_VAR];

#if FREERTOS_COPY_STACK
		// Stack wiederherstellen
		{
			portSTACK_TYPE *tmpPortSaveContextPtr;

			tmpPortSaveContextPtr = portSaveContextPtr;
			tmpPortSaveContextPtr -= REG_SAVECONTEXT_PTR;
			for (unsigned int i = (REG_CRITNEST_VAR + 1); i < (REG_SYS_NO - 0); i++)
				tmpPortSaveContextPtr[i] = taskSp[i];
		}
#else
		//die richtige Adresse fr den saveContextpointer eintragen
		portSaveContextPtr = &taskSp[REG_SAVECONTEXT_PTR];

		//da der gesamte Context in die Register berfhrt wird, knnten wir ihn ja schonmal nebenbei in den Cache holen
		//ob man es glaub oder nicht, das bringt einen Gewinn von 1,5 s! pro Taskwechsel
		CPU_prefetchContext(portSaveContextPtr);
#endif

#if FREERTOS_REENT_COPY_STRUCT
		{
			// Zeiger auf reent-struct des FreeRTOS-Tasks
			portSTACK_TYPE *pReent = (portSTACK_TYPE *)sysCall_getreent(CPU_IRQ_PRIOR_SYS);

			// wiederherstellen
			for (unsigned int i = 0; i < (REG_REENT_END - REG_REENT_START + 1); i++)
				pReent[i] = taskSp[REG_REENT_START + i];
		}
#elif FREERTOS_REENT_CHANGE_POINTER
		{
			// den globalen Zeiger auf das reent fr den Task wiederherstellen
			portCurrentTaskReentPtr = (struct _reent *)(((portSTACK_TYPE*)(taskSp))[REG_REENT_POINTER]);
		}
#endif

		//den aktuellen ContxtTCb merken, dass beim nchsten Restore auch wirklich dieser verwendet wird, um die Register des Tasks dorthin zu legen
		//sollte, das RTOS in der Zwischenzeit einen ContextSwitch machen, so wrde pxCurrentTCB gendert, und portSAVE_CONTEXT schreibtin den falschen Stack
		restoredTCB = pxCurrentTCB;

		NDEBUG_PIN_LOW_ (DEBUG_EDITOR_MBC, FREERTOS_DEBUG_PIN_RESTORE_CONTEXT);
		DEBUG_PIN_LOW_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_RESTORE_CONTEXT);
	}
}

int portSAVE_CONTEXT( void )
{
	//nur Umschalten, wenn notwendig
	if (portActAllowTaskSwitch == portSavedAllowTaskSwitch)
	{
		//verbieten, dass der Scheduler umschalten darf
		portSchedulerSuspended = TRUE;
		return FALSE;
	}
	else
	{
		//wenn Scheduler nicht umschalten durfte, dann wieder erlauben
		portSchedulerSuspended = FALSE;
		portSavedAllowTaskSwitch = portActAllowTaskSwitch;

		if (portSaveContextPtr != NULL)
		{
			portSTACK_TYPE *taskSp;

			DEBUG_PIN_HIGH_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SAVE_CONTEXT);
			NDEBUG_PIN_HIGH_ (DEBUG_EDITOR_MBC, FREERTOS_DEBUG_PIN_SAVE_CONTEXT);

#if FREERTOS_COPY_STACK
			//wenn im IRQ-/SWI-Handler die SystemModeRegister gesichert wurden, dann wird portSaveContextPtr gesetzt
			//und zeigt auf das REG_SAVECONTEXT_PTR -> den SP holen und an den Anfang setzen
			taskSp = (portSTACK_TYPE*)(portSaveContextPtr[REG_SYS_SP - REG_SAVECONTEXT_PTR]);
			taskSp -= REG_SYS_NO;

			// enterCritical/exitCritial-Zhler
			taskSp[REG_CRITNEST_VAR] = ulCriticalNesting;

			// aktuellen Stack sichern
			{
				portSTACK_TYPE *tmpPortSaveContextPtr;

				tmpPortSaveContextPtr = portSaveContextPtr;
				tmpPortSaveContextPtr -= REG_SAVECONTEXT_PTR;
				for (unsigned int i = (REG_CRITNEST_VAR + 1); i < (REG_SYS_NO - 0); i++)
					taskSp[i] = tmpPortSaveContextPtr[i];
			}
#else
			//wenn im IRQ-/SWI-Handler die SystemModeRegister gesichert wurden, dann wird portSaveContextPtr gesetzt
			//und zeigt auf das REG_SAVECONTEXT_PTR -> an den Anfang setzen
			taskSp = portSaveContextPtr;
			taskSp -= REG_SAVECONTEXT_PTR;

			// enterCritical/exitCritial-Zhler
			taskSp[REG_CRITNEST_VAR] = ulCriticalNesting;
#endif

#if FREERTOS_REENT_COPY_STRUCT
			{
				// Zeiger auf reent-struct des FreeRTOS-Tasks
				portSTACK_TYPE *pReent = (portSTACK_TYPE *)sysCall_getreent(CPU_IRQ_PRIOR_SYS);

				// sichern
				for (unsigned int i = 0; i < (REG_REENT_END - REG_REENT_START + 1); i++)
					taskSp[REG_REENT_START + i] = pReent[i];
			}
#elif FREERTOS_REENT_CHANGE_POINTER
			// Zeiger auf aktuelles reent-struct eintragen
			taskSp[REG_REENT_POINTER] = (portSTACK_TYPE)portCurrentTaskReentPtr;
#endif

#ifdef DBG_RTOSSTACKS_NO
			dbgRtosStacks[dbgRtosStackCnt].ticks		= portGET_RUN_TIME_COUNTER_VALUE();
			dbgRtosStacks[dbgRtosStackCnt].sp_sys		= (portSTACK_TYPE*)portSaveContextPtr[REG_SYS_SP - REG_SAVECONTEXT_PTR];
#if 0
			dbgRtosStacks[dbgRtosStackCnt].pc_sys		= (unsigned int)portSaveContextPtr[REG_IRQ_LR - REG_SAVECONTEXT_PTR];
			dbgRtosStacks[dbgRtosStackCnt].lr_sys		= (unsigned int)portSaveContextPtr[REG_SYS_LR - REG_SAVECONTEXT_PTR];
#elif 1
			dbgRtosStacks[dbgRtosStackCnt].pc_sys		= (unsigned int)portSaveContextPtr[REG_IRQ_LR - REG_SAVECONTEXT_PTR];
			dbgRtosStacks[dbgRtosStackCnt].lr_sys		= (void*)portSaveContextPtr[REG_SYS_LR - REG_SAVECONTEXT_PTR];
#else
			dbgRtosStacks[dbgRtosStackCnt].lr_sys		= (unsigned int)portSaveContextPtr[REG_SYS_LR - REG_SAVECONTEXT_PTR];
			dbgRtosStacks[dbgRtosStackCnt].pc_sys		= (void*)portSaveContextPtr[REG_IRQ_LR - REG_SAVECONTEXT_PTR];
#endif
			dbgRtosStacks[dbgRtosStackCnt].context		= portSaveContextPtr;
			dbgRtosStacks[dbgRtosStackCnt].TopOfStack	= taskSp;
			dbgRtosStacks[dbgRtosStackCnt].pxCurrentTCB	= (void *)restoredTCB;
			dbgRtosStacks[dbgRtosStackCnt].critNest		= ulCriticalNesting;
#ifdef DBG_IRQSTACKS_NO
			dbgRtosStacks[dbgRtosStackCnt].curDbgIrqStackCnt = dbgIrqStackCnt;
#endif

			// Inhalte des TCB
			if (dbgRtosStacks[dbgRtosStackCnt].pxCurrentTCB != NULL)
			{
				dbgRtosStacks[dbgRtosStackCnt].uxPriority		= dbgRtosStacks[dbgRtosStackCnt].pxCurrentTCB->uxPriority;
#if ( configGENERATE_RUN_TIME_STATS == 1 )
				dbgRtosStacks[dbgRtosStackCnt].ulRunTimeCounter	= dbgRtosStacks[dbgRtosStackCnt].pxCurrentTCB->ulRunTimeCounter;
#endif
			//	dbgRtosStacks[dbgRtosStackCnt].pcTaskName		= dbgRtosStacks[dbgRtosStackCnt].pxCurrentTCB->pcTaskName;
			}
			else
			{
				dbgRtosStacks[dbgRtosStackCnt].uxPriority		= 0;
#if ( configGENERATE_RUN_TIME_STATS == 1 )
				dbgRtosStacks[dbgRtosStackCnt].ulRunTimeCounter	= 0;
#endif
			//	dbgRtosStacks[dbgRtosStackCnt].pcTaskName		= NULL;
			}

			dbgRtosStackCnt++;
			if (dbgRtosStackCnt >= DBG_RTOSSTACKS_NO)
				dbgRtosStackCnt = 0;

			// den nchsten Eintrag lschen -> so findet man leicht auch den aktuellen Eintrag
			memset (&dbgRtosStacks[dbgRtosStackCnt], 0, sizeof(rtosStackDump_type));
#endif	// DBG_RTOSSTACKS_NO

#if FREE_RTOS_REG_STACK_ALIGNMENT
			// Fllbytes
			taskSp[REG_STACK_ALIGNMENT] = 0;
#endif
#if defined DEBUG_RTOS_STACK_CHECK_CRC
			// Prfsumme ber den aktuellen Stackinhalt
			taskSp[REG_STACK_CRC] = xPortCalcStackCrc (&taskSp[REG_STACK_CRC + 1]);
#endif

			//neuen TaskSP merken
			*restoredTCB = (portSTACK_TYPE)taskSp;

			NDEBUG_PIN_LOW_ (DEBUG_EDITOR_MBC, FREERTOS_DEBUG_PIN_SAVE_CONTEXT);
			DEBUG_PIN_LOW_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SAVE_CONTEXT);
		}
		return TRUE;
	}
}


/*-----------------------------------------------------------*/

void vPortISRStartFirstTask( void )
{
	//Scheduler starten
	//zunchst IRQs verbieten
	asm("	cpsid if");

	//ein paar Variablen setzen, so das das nchste Mal geswitched werden darf
	restoredTCB = pxCurrentTCB;
	assert (restoredTCB != NULL);

	portSavedAllowTaskSwitch	= 0;
	portActAllowTaskSwitch		= 1;
	portSchedulerSuspended		= FALSE;
	ulCriticalNesting			= ((portSTACK_TYPE*)(*restoredTCB))[REG_CRITNEST_VAR];
	portSaveContextPtr			= (portSTACK_TYPE*)(*restoredTCB) + REG_CRITNEST_VAR + 1;	// zeigt damit auf REG_SYS_FPEXC

#if FREERTOS_REENT_COPY_STRUCT
	{
		// Zeiger auf reent-struct des FreeRTOS-Tasks
		portSTACK_TYPE	*pStart = (portSTACK_TYPE*)(*restoredTCB) + REG_REENT_START;
		// Zeiger auf das globale reent-struct fr den IRQ-Level
		portSTACK_TYPE	*pReent = (portSTACK_TYPE *)sysCall_getreent(CPU_IRQ_PRIOR_SYS);

		// wiederherstellen
		for (unsigned int i = 0; i < (REG_REENT_END - REG_REENT_START + 1); i++)
			pReent[i] = pStart[i];
	}
#elif FREERTOS_REENT_CHANGE_POINTER
	// den globalen Zeiger auf das reent fr den Task wiederherstellen
	portCurrentTaskReentPtr = (struct _reent *)(((portSTACK_TYPE*)(*restoredTCB))[REG_REENT_POINTER]);
#endif

	// das FreeRTOS ruft vor dieser Funktion u.U. vPORT_DISABLE_IRQ,
	// beim Wiederherstellen des ersten Tasks werden diese allerdings automatisch
	// wieder freigegeben, allerdings nicht der Zhler, was zum Stillstand fhrt
	// Nach einem disableIrq/enableIrq-Paar ist der Zhler nicht 0 und damit
	// bleiben die IRQs gesperrt!
	// -> irqDisableNo zurcksetzen
	CPU_resetDisableIRQ ();

	//jetz alle Register zurckholen und raus hier
	FreeRTOS_restore1stContext();
}
/*-----------------------------------------------------------*/

/*
 * Called by portYIELD() or taskYIELD() to manually force a context switch.
 *
 * When a context switch is performed from the task level the saved task
 * context is made to look as if it occurred from within the tick ISR.  This
 * way the same restore context function can be used when restoring the context
 * saved from the ISR or that saved from a call to vPortYieldProcessor.
 */
void vPortYieldProcessor( void )
{
	/* Within an IRQ ISR the link register has an offset from the true return
	address, but an SWI ISR does not.  Add the offset manually so the same
	ISR return code can be used in both cases. */
//	asm volatile ( "ADD		LR, LR, #4" );

	DEBUG_PIN_HIGH_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_PROC);
	/* Perform the context switch.  First save the context of the current task. */
	if (portSAVE_CONTEXT())
	{
		/* Find the highest priority task that is ready to run. */
		DEBUG_PIN_LOW_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_PROC);
		vTaskSwitchContext();
		DEBUG_PIN_HIGH_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_PROC);
		/* Restore the context of the new task. */
		portRESTORE_CONTEXT();
	}
	DEBUG_PIN_LOW_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_PROC);
}

/*-----------------------------------------------------------*/

void vPortYield (void)
{
	DEBUG_PIN_HIGH_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD);
	// IRQ anmelden; fhrt vPortYieldProcessor aus nachdem die IRQs wieder freigegeben wurden
	CPU_setIRQIdx(CPU_IRQ_IDX_RTOS_YIELD);
	DEBUG_PIN_LOW_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD);
}

void vPortYield_ISR (void)
{
	DEBUG_PIN_HIGH_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_ISR);
	// IRQ anmelden; fhrt vPortYieldProcessor aus nachdem die IRQs wieder freigegeben wurden
	CPU_setIRQIdx(CPU_IRQ_IDX_RTOS_YIELD);
	DEBUG_PIN_LOW_ (DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_PORTYIELD_ISR);
}

/*-----------------------------------------------------------*/
/*
 * The ISR used for the scheduler tick depends on whether the cooperative or
 * the preemptive scheduler is being used.
 */

#if configUSE_PREEMPTION == 0	// FIXME: broken

	/* The cooperative scheduler requires a normal IRQ service routine to
	simply increment the system tick. */
	void vNonPreemptiveTick( void );// __attribute__ ((interrupt ("IRQ")));
	void vNonPreemptiveTick( void )
	{
		unsigned long ulDummy __attribute__((__unused__));

		/* Increment the tick count - which may wake some tasks but as the
		preemptive scheduler is not being used any woken task is not given
		processor time no matter what its priority. */
		vTaskIncrementTick();

		/* Clear the PIT interrupt. */
		ulDummy = PIT->PIT_PIVR;

	}

#else

	unsigned int (*portTickFunc)(void) = NULL;

	/* The preemptive scheduler is defined as "naked" as the full context is
	saved on entry as part of the context switch. */
	void vPreemptiveTick( void );// __attribute__((naked));
	void vPreemptiveTick( void )
	{
		DEBUG_PIN_HIGH_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);

		/* Save the context of the current task. */
		if (portSAVE_CONTEXT())
		{
			unsigned int ticks;

			if (portTickFunc)
			{
				DEBUG_PIN_TOGGLE_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);
				ticks = portTickFunc();
				DEBUG_PIN_TOGGLE_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);
			}
			else
			{
				ticks = 1;
			}

			while (ticks--)
			{
				/* Increment the tick count - this may wake a task. */
				vTaskIncrementTick();
			}

			/* Find the highest priority task that is ready to run. */
			DEBUG_PIN_TOGGLE_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);
			vTaskSwitchContext();
			DEBUG_PIN_TOGGLE_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);

			portRESTORE_CONTEXT();
		}
		else
		{
			unsigned int ticks;

			if (portTickFunc)
				ticks = portTickFunc();
			else
				ticks = 1;

			while (ticks--)
			{
				/* Increment the tick count - this may wake a task. */
				vTaskIncrementTick();
			}
		}
		DEBUG_PIN_LOW_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_SYSTICK);
	}

	void vPortWriteTickFunc( unsigned int (*func)(void))
	{
		portTickFunc = func;
	}

#endif
/*-----------------------------------------------------------*/

void vPortDisableInterrupts (void)
{
	CPU_disableIRQ0();
}
/*-----------------------------------------------------------*/

void vPortEnableInterrupts (void)
{
	CPU_enableIRQ0();
}
/*-----------------------------------------------------------*/

#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
const void *vPortEnterCriticalCaller;
const void *vPortExitCriticalCaller;
#endif

/* The code generated by the GCC compiler uses the stack in different ways at
different optimisation levels.  The interrupt flags can therefore not always
be saved to the stack.  Instead the critical section nesting level is stored
in a variable, which is then saved as part of the stack context. */
#ifdef SYS_CPU_INTRAM_TEXT_SECTION
void vPortEnterCritical( void )	__attribute__ ((section (SYS_CPU_INTRAM_TEXT_SECTION),hot));
#endif
void vPortEnterCritical( void )
{
	DEBUG_PIN_HIGH_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_ENTERCRITICAL);

#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
	unsigned long lr;
	// Linkregister sichern
	asm(" mov	%0, lr" : "=r" (lr));

	vPortEnterCriticalCaller = (void *)lr;
	vPortExitCriticalCaller = NULL;
#endif

	//wir nehmen einfach unsere altbekannte Funktion
	CPU_disableIRQ0();

	/* Now interrupts are disabled ulCriticalNesting can be accessed
	directly.  Increment ulCriticalNesting to keep a count of how many times
	portENTER_CRITICAL() has been called. */
	if( ulCriticalNesting > portNO_CRITICAL_NESTING )
	{
		// ein weiterer Aufruf von vPortEnterCritical
		ulCriticalNesting++;

		// weil hier CPU_disableIRQ0 immer gerufen wird,
		// im vPortExitCritical das CPU_enableIRQ0 aber erst nur einmal ganz am Schluss
		// muss man hier die CPU_irq.disableNo wieder um eins erniedrigen,
		// durch einen Aufruf von CPU_enableIRQ,
		// der allerdings die IRQs selbst noch nicht freigeben sollte,
		// weil der Zhler dort auch nach dem enable noch grer als 0 ist
		CPU_enableIRQ0();
	}
	else
	{
		// das war der allerste Aufruf -> die IRQs bleiben verboten
		ulCriticalNesting++;
	}
	DEBUG_PIN_LOW_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_ENTERCRITICAL);
}

#ifdef SYS_CPU_INTRAM_TEXT_SECTION
void vPortExitCritical( void )	__attribute__ ((section (SYS_CPU_INTRAM_TEXT_SECTION),hot));
#endif
void vPortExitCritical( void )
{
	DEBUG_PIN_HIGH_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_EXITCRITICAL);
#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
	unsigned long lr;
	// Linkregister sichern
	asm(" mov	%0, lr" : "=r" (lr));

	vPortExitCriticalCaller = (void *)lr;
#endif

	if( ulCriticalNesting > portNO_CRITICAL_NESTING )
	{
		/* Decrement the nesting count as we are leaving a critical section. */
		ulCriticalNesting--;

		/* If the nesting level has reached zero then interrupts should be
		re-enabled. */
		if( ulCriticalNesting == portNO_CRITICAL_NESTING )
		{
			//wir nehmen einfach unsere altbekannte Funktion
			CPU_enableIRQ0();
		}
		else
		{
			// zum Brechen
			asm volatile ("nop");
		}
	}
	else
	{
		// zum Brechen
		asm volatile ("nop");
	}
	DEBUG_PIN_LOW_(DEBUG_EDITOR_AHC, FREERTOS_DEBUG_PIN_EXITCRITICAL);
}

