/*
    FreeRTOS V7.4.2 - Copyright (C) 2013 Real Time Engineers Ltd.

    FEATURES AND PORTS ARE ADDED TO FREERTOS ALL THE TIME.  PLEASE VISIT
    http://www.FreeRTOS.org TO ENSURE YOU ARE USING THE LATEST VERSION.

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.

    >>>>>>NOTE<<<<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.

    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details. You should have received a copy of the GNU General Public License
    and the FreeRTOS license exception along with FreeRTOS; if not it can be
    viewed here: http://www.freertos.org/a00114.html and also obtained by
    writing to Real Time Engineers Ltd., contact details for whom are available
    on the FreeRTOS WEB site.

    1 tab == 4 spaces!

    ***************************************************************************
     *                                                                       *
     *    Having a problem?  Start by reading the FAQ "My application does   *
     *    not run, what could be wrong?"                                     *
     *                                                                       *
     *    http://www.FreeRTOS.org/FAQHelp.html                               *
     *                                                                       *
    ***************************************************************************


    http://www.FreeRTOS.org - Documentation, books, training, latest versions,
    license and Real Time Engineers Ltd. contact details.

    http://www.FreeRTOS.org/plus - A selection of FreeRTOS ecosystem products,
    including FreeRTOS+Trace - an indispensable productivity tool, and our new
    fully thread aware and reentrant UDP/IP stack.

    http://www.OpenRTOS.com - Real Time Engineers ltd license FreeRTOS to High
    Integrity Systems, who sell the code with commercial support,
    indemnification and middleware, under the OpenRTOS brand.

    http://www.SafeRTOS.com - High Integrity Systems also provide a safety
    engineered and independently SIL3 certified version for use in safety and
    mission critical applications that require provable dependability.
*/

/*
	Changes from V3.2.3

	+ Modified portENTER_SWITCHING_ISR() to allow use with GCC V4.0.1.

	Changes from V3.2.4

	+ Removed the use of the %0 parameter within the assembler macros and
	  replaced them with hard coded registers.  This will ensure the
	  assembler does not select the link register as the temp register as
	  was occasionally happening previously.

	+ The assembler statements are now included in a single asm block rather
	  than each line having its own asm block.

	Changes from V4.5.0

	+ Removed the portENTER_SWITCHING_ISR() and portEXIT_SWITCHING_ISR() macros
	  and replaced them with portYIELD_FROM_ISR() macro.  Application code
	  should now make use of the portSAVE_CONTEXT() and portRESTORE_CONTEXT()
	  macros as per the V4.5.1 demo code.
*/

#ifndef PORTMACRO_H
#define PORTMACRO_H

#ifdef __cplusplus
extern "C" {
#endif

#include <debug.h>

/*-----------------------------------------------------------
 * Port specific definitions.
 *
 * The settings in this file configure FreeRTOS correctly for the
 * given hardware and compiler.
 *
 * These settings should not be altered.
 *-----------------------------------------------------------
 */

/* Type definitions. */
#define portCHAR		char
#define portFLOAT		float
#define portDOUBLE		double
#define portLONG		long
#define portSHORT		short
#define portSTACK_TYPE	unsigned portLONG
#define portBASE_TYPE	portLONG

#if( configUSE_16_BIT_TICKS == 1 )
	typedef unsigned portSHORT portTickType;
	#define portMAX_DELAY ( portTickType ) 0xffff
#else
	typedef unsigned portLONG portTickType;
	#define portMAX_DELAY ( portTickType ) 0xffffffff
#endif
/*-----------------------------------------------------------*/

/* Architecture specifics. */
#define portSTACK_GROWTH			( -1 )
#define portTICK_RATE_MS			( ( portTickType ) 1000 / configTICK_RATE_HZ )
#define portBYTE_ALIGNMENT			8
#define portNOP()					asm volatile ( "NOP" );
/*-----------------------------------------------------------*/

/* Scheduler utilities. */

#if 0
/*
 * portRESTORE_CONTEXT, portRESTORE_CONTEXT, portENTER_SWITCHING_ISR
 * and portEXIT_SWITCHING_ISR can only be called from ARM mode, but
 * are included here for efficiency.  An attempt to call one from
 * THUMB mode code will result in a compile time error.
 */

#define portRESTORE_CONTEXT_norm()										\
{																		\
extern portSTACK_TYPE *pxCurrentTCB;							\
extern volatile unsigned portLONG ulCriticalNesting;					\
																		\
	/* Set the LR to the task stack. */									\
	asm volatile (														\
	"LDR		R0, =pxCurrentTCB								\n\t"	\
	"LDR		R0, [R0]										\n\t"	\
	"LDR		LR, [R0]										\n\t"	\
																		\
	/* The critical nesting depth is the first item on the stack. */	\
	/* Load it into the ulCriticalNesting variable. */					\
	"LDR		R0, =ulCriticalNesting							\n\t"	\
	"LDMFD	LR!, {R1}											\n\t"	\
	"STR		R1, [R0]										\n\t"	\
																		\
	/* Get the SPSR from the stack. */									\
	"LDMFD	LR!, {R0}											\n\t"	\
	"MSR		SPSR, R0										\n\t"	\
																		\
	/* Restore all system mode registers for the task. */				\
	"LDMFD	LR, {R0-R14}^										\n\t"	\
	"NOP														\n\t"	\
																		\
	/* Restore the return address. */									\
	"LDR		LR, [LR, #+60]									\n\t"	\
																		\
	/* And return - correcting the offset in the LR to obtain the */	\
	/* correct address. */												\
	"SUBS	PC, LR, #4											\n\t"	\
	);																	\
	( void ) ulCriticalNesting;											\
	( void ) pxCurrentTCB;												\
}
/*-----------------------------------------------------------*/

#define portSAVE_CONTEXT_norm()												\
{																		\
extern portSTACK_TYPE *pxCurrentTCB;							\
extern volatile unsigned portLONG ulCriticalNesting;					\
																		\
	/* Push R0 as we are going to use the register. */					\
	asm volatile (														\
	"STMDB	SP!, {R0}											\n\t"	\
																		\
	/* Set R0 to point to the task stack pointer. */					\
	"STMDB	SP,{SP}^											\n\t"	\
	"NOP														\n\t"	\
	"SUB	SP, SP, #4											\n\t"	\
	"LDMIA	SP!,{R0}											\n\t"	\
																		\
	/* Push the return address onto the stack. */						\
	"STMDB	R0!, {LR}											\n\t"	\
																		\
	/* Now we have saved LR we can use it instead of R0. */				\
	"MOV	LR, R0												\n\t"	\
																		\
	/* Pop R0 so we can save it onto the system mode stack. */			\
	"LDMIA	SP!, {R0}											\n\t"	\
																		\
	/* Push all the system mode registers onto the task stack. */		\
	"STMDB	LR,{R0-LR}^											\n\t"	\
	"NOP														\n\t"	\
	"SUB	LR, LR, #60											\n\t"	\
																		\
	/* Push the SPSR onto the task stack. */							\
	"MRS	R0, SPSR											\n\t"	\
	"STMDB	LR!, {R0}											\n\t"	\
																		\
	"LDR	R0, =ulCriticalNesting								\n\t"	\
	"LDR	R0, [R0]											\n\t"	\
	"STMDB	LR!, {R0}											\n\t"	\
																		\
	/* Store the new top of stack for the task. */						\
	"LDR	R0, =pxCurrentTCB									\n\t"	\
	"LDR	R0, [R0]											\n\t"	\
	"STR	LR, [R0]											\n\t"	\
	);																	\
	( void ) ulCriticalNesting;											\
	( void ) pxCurrentTCB;												\
}
#else
extern int portSAVE_CONTEXT( void );
extern void portRESTORE_CONTEXT( void );
#endif

#if 0
#define portYIELD()													\
	asm volatile ( "stmdb	sp!, {lr}		//LR sichern	" );	\
	asm volatile ( "SWI 0									" );	\
	asm volatile ( "ldmia	sp!, {lr}		//LR holen		" );
#else
extern void vPortYield (void);
extern void vPortYield_ISR (void);

#define portYIELD()					vPortYield()
#define portYIELD_FROM_ISR()		vPortYield_ISR()
#endif

void vPortYieldProcessor( void );

void vPreemptiveTick( void );

void vPortWriteTickFunc( unsigned int (*func)(void));


/*-----------------------------------------------------------*/


/* Critical section management. */

/*
 * The interrupt management utilities can only be called from ARM mode.  When
 * THUMB_INTERWORK is defined the utilities are defined as functions in
 * portISR.c to ensure a switch to ARM mode.  When THUMB_INTERWORK is not
 * defined then the utilities are defined as macros here - as per other ports.
 */

#ifdef THUMB_INTERWORK

	extern void vPortDisableInterruptsFromThumb( void ) __attribute__ ((naked));
	extern void vPortEnableInterruptsFromThumb( void ) __attribute__ ((naked));

	#define portDISABLE_INTERRUPTS()	vPortDisableInterruptsFromThumb()
	#define portENABLE_INTERRUPTS()		vPortEnableInterruptsFromThumb()

#else
#if 1
extern void vPortDisableInterrupts (void);
extern void vPortEnableInterrupts (void);

#define portDISABLE_INTERRUPTS()	vPortDisableInterrupts()
#define portENABLE_INTERRUPTS()		vPortEnableInterrupts()

#else
	#define portDISABLE_INTERRUPTS()											\
		asm volatile (															\
			"STMDB	SP!, {R0}		\n\t"	/* Push R0.						*/	\
			"MRS	R0, CPSR		\n\t"	/* Get CPSR.					*/	\
			"ORR	R0, R0, #0xC0	\n\t"	/* Disable IRQ, FIQ.			*/	\
			"MSR	CPSR, R0		\n\t"	/* Write back modified value.	*/	\
			"LDMIA	SP!, {R0}			" )	/* Pop R0.						*/

	#define portENABLE_INTERRUPTS()												\
		asm volatile (															\
			"STMDB	SP!, {R0}		\n\t"	/* Push R0.						*/	\
			"MRS	R0, CPSR		\n\t"	/* Get CPSR.					*/	\
			"BIC	R0, R0, #0xC0	\n\t"	/* Enable IRQ, FIQ.				*/	\
			"MSR	CPSR, R0		\n\t"	/* Write back modified value.	*/	\
			"LDMIA	SP!, {R0}			" )	/* Pop R0.						*/
#endif
#endif /* THUMB_INTERWORK */

extern void vPortEnterCritical( void );
extern void vPortExitCritical( void );

#define portENTER_CRITICAL()		vPortEnterCritical()
#define portEXIT_CRITICAL()			vPortExitCritical()
/*-----------------------------------------------------------*/

/* Task function macros as described on the FreeRTOS.org WEB site. */
#define portTASK_FUNCTION_PROTO( vFunction, pvParameters ) void vFunction( void *pvParameters )
#define portTASK_FUNCTION( vFunction, pvParameters ) void vFunction( void *pvParameters )

/*-----------------------------------------------------------*/

/* User defined macros for priority handling required,
 * because xSemaphoreGiveFromISR (and others) may corrupt uxTopReadyPriority
 * if interrupted by vPreemptiveTick!
 */
/* uxTopReadyPriority holds the priority of the highest priority ready state task. */
#define portRECORD_READY_PRIORITY( uxPriority, uxTopReadyPriority )													\
do {																												\
	/* Critical Section required, access to uxTopReadyPriority must be interrupt safe! */							\
	portENTER_CRITICAL();																							\
	if( ( uxPriority ) > uxTopReadyPriority )																		\
	{																												\
		uxTopReadyPriority = ( uxPriority );																		\
	}																												\
	portEXIT_CRITICAL();																							\
} while(0) /* taskRECORD_READY_PRIORITY */


/* Find the highest priority queue that contains ready tasks. */
#if 0	// hilft auch nicht, TMC blieb immer noch hngen
// es wird der Ready-Task mit der hchsten Prio gesucht und sich die Prio des Treffers gemerkt fr die nchste Suche
#define portGET_HIGHEST_PRIORITY( uxTopPriority, uxTopReadyPriority )												\
do {																												\
	/* Critical Section required, access to uxTopReadyPriority must be interrupt safe! */							\
	portENTER_CRITICAL();																							\
	/* Find the highest priority queue that contains ready tasks. */												\
	while( listLIST_IS_EMPTY( &( pxReadyTasksLists[ uxTopReadyPriority ] ) ) )										\
	{																												\
		configASSERT( uxTopReadyPriority );																			\
		--uxTopReadyPriority;																						\
	}																												\
	uxTopPriority = uxTopReadyPriority;																				\
	portEXIT_CRITICAL();																							\
} while(0)
#else
// es wird der Ready-Task mit der hchsten Prio gesucht, aber die Prio des Treffers NICHT gemerkt
// diese Variante dauert etwas lnger, weil immer alle (benutzten) Prioritten durchsucht werden
#define portGET_HIGHEST_PRIORITY( uxTopPriority, uxTopReadyPriority )												\
do {																												\
	/* Critical Section required, access to uxTopReadyPriority must be interrupt safe! */							\
	portENTER_CRITICAL();																							\
	uxTopPriority = uxTopReadyPriority;																				\
	portEXIT_CRITICAL();																							\
	/* Find the highest priority queue that contains ready tasks. */												\
	while( listLIST_IS_EMPTY( &( pxReadyTasksLists[ uxTopPriority ] ) ) )											\
	{																												\
		configASSERT( uxTopPriority );																				\
		--uxTopPriority;																							\
	}																												\
} while(0)
#endif

/* Default implementation is empty */
#define portRESET_READY_PRIORITY( uxPriority, uxTopReadyPriority )

/*-----------------------------------------------------------*/

/* Critical Section erforderlich:
 * Wenn innerhalb von xSemaphoreGiveFromISR der vPreemptiveTick drankommt
 * und den Task freigibt weil die Blockzeit abgelaufen ist,
 * dann landet man u.U. im DataAbort, weil der Zeiger auf den wartenden Task in der Queue pltzlich (und ungetestet) NULL ist! */
#define portSemaphoreGiveFromISR( xSemaphore, pxHigherPriorityTaskWoken )											\
do {																												\
	portENTER_CRITICAL();																							\
	xSemaphoreGiveFromISR( xSemaphore, pxHigherPriorityTaskWoken);													\
	portEXIT_CRITICAL();																							\
} while(0)

/*-----------------------------------------------------------*/

// FIXME: das braucht man fr die Runtime stats
#define portCONFIGURE_TIMER_FOR_RUN_TIME_STATS()		(void)0
#define portGET_RUN_TIME_COUNTER_VALUE()				0

#include <reent.h>

// Anzahl der "Register" fr die reent-Struktur
#define FREE_RTOS_REENT_SIZE				((sizeof(struct _reent) + sizeof(portSTACK_TYPE) - 1) / sizeof(portSTACK_TYPE))
#ifdef FREE_RTOS_REENT_SIZE
	//wir wollen, dass jedesmal beim Taskwechsel die reent-Struktur kopiert wird?
	#if defined DEBUG_EDITOR && ((defined DEBUG_EDITOR_AHC && DEBUG_EDITOR == DEBUG_EDITOR_AHC)	\
							  || (defined DEBUG_EDITOR_MBC && DEBUG_EDITOR == DEBUG_EDITOR_MBC))
		//n, wollen wir nich
		#define FREERTOS_REENT_COPY_STRUCT			FALSE
	#else
		#define FREERTOS_REENT_COPY_STRUCT			TRUE
	#endif
	#if FREERTOS_REENT_COPY_STRUCT == FALSE
		//besser, die reent Struktur liegt einmal an Ende des Stacks des Task beim Init und nur der Zeiger wird getauscht
		#define FREERTOS_REENT_CHANGE_POINTER		TRUE
		#define FREE_RTOS_REG_STACK_ALIGNMENT		FALSE
	#endif

	#if FREERTOS_REENT_CHANGE_POINTER
		// Zeiger auf das reent-struct im Stack des aktuellen Task
		extern struct _reent *portCurrentTaskReentPtr;
	#endif
#endif

#define FREERTOS_COPY_STACK							TRUE	//wenn TRUE, den alten irqHandler in _isrPort.s nehmen (dort FREERTOS_COPY_STACK auf 1 setzen)

// braucht man oder nicht, merkt man am assert...
#ifndef FREE_RTOS_REG_STACK_ALIGNMENT
	#define FREE_RTOS_REG_STACK_ALIGNMENT			TRUE
#endif

enum contextStruct
{
	REG_STACK_CRC = 0,
#if FREE_RTOS_REG_STACK_ALIGNMENT
	REG_STACK_ALIGNMENT,
#endif
#if FREERTOS_REENT_COPY_STRUCT
	REG_REENT_START,
	REG_REENT_END = REG_REENT_START + FREE_RTOS_REENT_SIZE - 1,
#elif FREERTOS_REENT_CHANGE_POINTER
	REG_REENT_POINTER,
#endif
	REG_CRITNEST_VAR,
	REG_SYS_FPEXC,
	REG_SYS_FPSCR,
	REG_SYS_S00,
	REG_SYS_S01,
	REG_SYS_S02,
	REG_SYS_S03,
	REG_SYS_S04,
	REG_SYS_S05,
	REG_SYS_S06,
	REG_SYS_S07,
	REG_SYS_S08,
	REG_SYS_S09,
	REG_SYS_S10,
	REG_SYS_S11,
	REG_SYS_S12,
	REG_SYS_S13,
	REG_SYS_S14,
	REG_SYS_S15,
	REG_SYS_S16,
	REG_SYS_S17,
	REG_SYS_S18,
	REG_SYS_S19,
	REG_SYS_S20,
	REG_SYS_S21,
	REG_SYS_S22,
	REG_SYS_S23,
	REG_SYS_S24,
	REG_SYS_S25,
	REG_SYS_S26,
	REG_SYS_S27,
	REG_SYS_S28,
	REG_SYS_S29,
	REG_SYS_S30,
	REG_SYS_S31,
	REG_SYS_CPSR,
	REG_SYS_R0,
	REG_SYS_R1,
	REG_SYS_R2,
	REG_SYS_R3,
	REG_SYS_R4,
	REG_SYS_R5,
	REG_SYS_R6,
	REG_SYS_R7,
	REG_SYS_R8,
	REG_SYS_R9,
	REG_SYS_R10,
	REG_SYS_R11,
	REG_SYS_R12,
	REG_SYS_SP,
	REG_SYS_LR,
	REG_IRQ_LR,
	REG_SYS_NO,
#if FREERTOS_COPY_STACK
	REG_SAVECONTEXT_PTR = REG_SYS_R0
#else
	REG_SAVECONTEXT_PTR = REG_SYS_FPEXC
#endif
};

#ifdef __cplusplus
}
#endif

#endif /* PORTMACRO_H */

