/*==========================================================================
	Datei:	mem_alloc.h

	ffentliche Header Datei fr Malloc-berwachung

	Begonnen:			22.08.2010
	Letzte nderung:	22.08.2010
	Autor:				M.Bhme\HAMEG Chemnitz
===========================================================================*/
#ifndef MEM_ALLOC_H
#define MEM_ALLOC_H

#ifdef __cplusplus
extern "C" {
#endif

/*--- Includes ------------------------------------------------------------*/
#include <stddef.h>
#include <stdint.h>

/*--- Makros & Konstanten -------------------------------------------------*/

// Kontrolle von malloc & Co.: welches Modul allokiert was
enum MEM_ALLOC_MODULE_LIST
{
	MEM_ALLOC_MODULE_NANO_X = 0,		/* 0, immer das erste, weil dort einfach 0 bergeben wird */
	MEM_ALLOC_MODULE_FREETYPE,			/* 1 */

	// MesOS-Module
	MEM_ALLOC_MODULE_DBG_CONSOLE,		/* 2 */
	MEM_ALLOC_MODULE_DS,				/* 3 */
	MEM_ALLOC_MODULE_SPI,				/* 4 */
	MEM_ALLOC_MODULE_FS,				/* 5 */
	MEM_ALLOC_MODULE_IMAGE,				/* 6 */
	MEM_ALLOC_MODULE_GZIP,				/* 7 */
	MEM_ALLOC_MODULE_MEM,				/* 8 */
	MEM_ALLOC_MODULE_RTOS,				/* 9 */
	MEM_ALLOC_MODULE_FPGA,				/* 10 */
	MEM_ALLOC_MODULE_LWIP,				/* 11 */
	MEM_ALLOC_MODULE_USB,				/* 12 */

	// eigentlich nicht der richtige Ort: HCL-Module
	MEM_ALLOC_MODULE_DDF,				/* 13 */
	MEM_ALLOC_MODULE_DATA_IO,			/* 14 */
	MEM_ALLOC_MODULE_STORE,				/* 15 */
	MEM_ALLOC_MODULE_STORE_DB,			/* 16 */
	MEM_ALLOC_MODULE_SCPI,				/* 17 */
	MEM_ALLOC_MODULE_UPDATE_PLD,		/* 18 */

	// HMF
	MEM_ALLOC_MODULE_HELP,				/* 19 */
	MEM_ALLOC_MODULE_ARBITRARY,			/* 20 */
	MEM_ALLOC_MODULE_ARBITRARY_POINTS,	/* 21 */
	MEM_ALLOC_MODULE_ARBITRARY_STORAGE,	/* 22 */
	MEM_ALLOC_MODULE_GRAPHIC_SIGNALS,	/* 23 */
	MEM_ALLOC_MODULE_GRAPHIC_MENU,		/* 24 */
	MEM_ALLOC_MODULE_MESSAGE,			/* 25 */
	MEM_ALLOC_MODULE_SWEEP,				/* 26 */
	MEM_ALLOC_MODULE_ADD_ON,			/* 27 */
	MEM_ALLOC_MODULE_MATH,				/* 28 */
	MEM_ALLOC_MODULE_MENU,				/* 39 */
	MEM_ALLOC_MODULE_PARAMETER,			/* 40 */

	MEM_ALLOC_MODULE_ADJUST,			/* 41 */

	MEM_ALLOC_MODULE_NO					/* 42 */
};

/* Wenn man an der berwachung von malloc & Co. interessiert ist,
 * dann sollte man die entsprechenden #defines in seine debug_pr.h
 * eintragen, nicht hier!
 */

/* malloc & Co. berwachen
 * fhrt Buch ber die Anzahlen der Funktionsrufe und die allokierte Speichermenge aller Module
 * individuell in mem_allocModules
 * insgesamt in mem_allocTotal
 */
//#define MEM_ALLOC_DEBUG

/* Schreibaktion vor und/oder hinter den allokierten Speicherblcken berwachen
 * in einer Debug-bersetzung gibts einen assert bei Unstimmigkeiten
 */
//#define MEM_ALLOC_DEBUG_FENCE

/* wenn man mal einen genauen Blick auf ein bestimmtes Modul werfen will
 * trgt alle allokierten Speicherblcke in die mem_allocTable ein (und auch wieder aus)
 * wahlweise eines einzelnen Moduls durch Angabe der entsprechenden ID
 * oder gleich aller Module durch Angabe von MEM_ALLOC_MODULE_NO
 */
//#define MEM_ALLOC_LOG_MODULE_ID		MEM_ALLOC_MODULE_NANO_X
//#define MEM_ALLOC_LOG_MODULE_ID		MEM_ALLOC_MODULE_NO


// Anzahl der Eintrge in der berwachungstabelle
//#ifndef MEM_ALLOC_LOG_TABLE_SIZE
//#define MEM_ALLOC_LOG_TABLE_SIZE		512
//#endif

//#define MEM_ALLOC_TABLE_ENTRY_USED		(1<<0)

/*--- Type Definitionen ---------------------------------------------------*/

// Informationen zu einem Modul
struct mem_allocModuleInfo
{
	int32_t		mallocCnt;		/*!< Anzahl der Aufrufe von malloc */
	int32_t		reallocCnt;		/*!< Anzahl der Aufrufe von realloc */
	int32_t		freeCnt;		/*!< Anzahl der Aufrufe von free */
	int32_t		totalMem;		/*!< allokierter Speicher des Moduls */
	int32_t		active;			/*!< Module allokiert gerade */
};

#ifdef MEM_ALLOC_DEBUG
// damits der Debugger einfach anzeigen kann
struct mem_allocTableFence
{
	uint8_t		d[8];			/*!< unbedingt das Alignment auf 8 Byte erhalten! */
};

#ifdef MEM_ALLOC_LOG_MODULE_ID
// Eintrag in der Speicherblocktabelle
struct mem_allocTableEntry
{
	struct mem_allocTableEntry			*next;	/*!< verkette Liste */
	const uint8_t						*mm;	/*!< Startzeiger auf allokierten Speicherblock */
	size_t								sz;		/*!< Gre des allokierten Speicherblocks */
//	uint8								st;		/*!< Status: Eintrag verwendet, etc. */

#ifdef MEM_ALLOC_DEBUG_FENCE
	const struct mem_allocTableFence	*he;	/*!< Zeiger auf den Zaun vor dem allokierten Speicher */
	const struct mem_allocTableFence	*ta;	/*!< Zeiger auf den Zaun nach dem allokierten Speicher */
#endif

	enum MEM_ALLOC_MODULE_LIST			id;		/*!< welches Module hat den Speicher allokiert, als enum, damit der Debugger es schn im Klartext anzeigt */
};

#endif
#endif

/*--- Globals -------------------------------------------------------------*/

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
// Gesamtstatistik
extern struct mem_allocModuleInfo	mem_allocTotal;
#endif

#ifdef MEM_ALLOC_DEBUG
// berwachung der einzelnen Module
extern struct mem_allocModuleInfo	mem_allocModules[MEM_ALLOC_MODULE_NO];

#ifdef MEM_ALLOC_LOG_MODULE_ID
// Tabelle mit allen durch malloc, calloc und realloc allokierten Speicherblcken eines oder aller Module
extern struct mem_allocTableEntry	mem_allocTable[/*MEM_ALLOC_LOG_TABLE_SIZE*/];
extern const uint32_t				mem_allocTableSize;

#endif
#endif

/*--- Prototypen ----------------------------------------------------------*/

/*!	\fn void * mem_alloc (size_t size, unsigned int moduleId)
	\brief allokiert Speicherblock mit Gre size
	\param size gewnschte Gre des Speicherblocks in Byte
	\param moduleId Eintrag aus der Liste MEM_ALLOC_MODULE_LIST
	\return NULL bei Fehler, sonst Zeiger auf Speicherblock
*/
void * mem_alloc (size_t size, unsigned int moduleId);

/*!	\fn void * mem_calloc (size_t elems, size_t elemSize, unsigned int moduleId)
	\brief allokiert Speicherblock fr elems Elemente mit Gre elemSize
	\param elems Anzahl der Elemente im Speicherblock
	\param elemSize Gre eines Elements in Byte
	\param moduleId Eintrag aus der Liste MEM_ALLOC_MODULE_LIST
	\return NULL bei Fehler, sonst Zeiger auf Speicherblock
*/
void * mem_calloc (size_t elems, size_t elemSize, unsigned int moduleId);

/*!	\fn void mem_free (void * memPtr, size_t size, unsigned int moduleId)
	\brief gibt allokierten Speicherblock frei
	\param memPtr Zeiger auf allokierten Speicherblock
	\param size soweit bekannt die Gre des Speicherblocks in Byte, oder 0
	\param moduleId Eintrag aus der Liste MEM_ALLOC_MODULE_LIST
	\return Nichts
*/
void mem_free (void * memPtr, size_t size, unsigned int moduleId);

/*!	\fn void * mem_realloc (void * memPtr, size_t oldSize, size_t newSize, unsigned int moduleId)
	\brief ndert die Gre eines allokierten Speicherblocks
	\param memPtr Zeiger auf allokierten Speicherblock
	\param oldSize soweit bekannt die aktuelle Gre des Speicherblocks in Byte, oder 0
	\param newSize gewnschte neue Gre des Speicherblocks in Byte
	\param moduleId Eintrag aus der Liste MEM_ALLOC_MODULE_LIST
	\return NULL bei Fehler, sonst Zeiger auf Speicherblock
*/
void * mem_realloc (void * memPtr, size_t oldSize, size_t newSize, unsigned int moduleId);

#ifdef __cplusplus
}
#endif
#endif	// MEM_ALLOC_H
